import os
import sys
sys.path.append('../almaqso')  # Specify the path to the almaqso directory
from almaqso.download_archive import download_archive
from almaqso.analysis import analysis


if __name__ == '__main__':
    download_archive(4, 'catalog/test_2.json')
    analysis('.', casacmd='/usr/local/casa/casa-6.6.1-17-pipeline-2024.1.0.8/bin/casa')  # Specify the path to the CASA binary



import os
import sys
from pathlib import Path
sys.path.append('../almaqso')  # isort:skip
from almaqso.download_archive import download_archive
from almaqso.analysis import analysis

CURRENT_DIR = Path(__file__).parent
CASA_PATH = 'casa'  # Path to CASA. Please modify this if necessary.
DATA_PATH = './uid___A002_Xd68367_X9885'
TARFILE = './2018.1.01575.S_uid___A002_Xd68367_X9885.asdm.sdm.tar'

# Edit the following constants.
DOWNLOAD = False  # True: Download the tar file, False: Use the existing tar file


def test_download():
    if DOWNLOAD:
        os.system(f'rm -rf {TARFILE}')
        download_archive(4, 'catalog/test_2.json')
        assert os.path.exists(TARFILE)


def test_analysis():
    os.system(f'rm -rf {DATA_PATH}')

    analysis('.', casacmd=CASA_PATH)

    os.system(f'rm -rf {DATA_PATH}_backup')
    os.system(f'cp -r {DATA_PATH} {DATA_PATH}_backup')



import os
from .QSOanalysis import QSOanalysis


def _analysis(tarfilename: str, skip=True, casacmd='casa'):
    obj = QSOanalysis(tarfilename, spacesave=True, casacmd=casacmd, casacmdforuvfit='casa')

    asdmname = 'uid___' + (tarfilename.split('_uid___')
                           [1]).replace('.asdm.sdm.tar', '')

    if os.path.exists(asdmname) and skip:
        print(asdmname+': analysis already done and skip')

    else:
        if os.path.exists(asdmname):
            print(asdmname+': analysis already done but reanalyzed')

        print('step:0')
        obj.intial_proc()
        print('step:1')
        obj.importasdm()
        print('step:2')
        obj.gen_calib_script()
        print('step:3')
        obj.remove_target()
        print('step:4')
        obj.doCalib()



import os
import subprocess


def _run_casa_script(casa: str, cmd: str, cmd_name: str):
    try:
        result = subprocess.run(
            [casa, '--nologger', '--nogui', '-c', cmd],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )
        print(f"STDOUT for {cmd_name}:", result.stdout)
        print(f"STDERR for {cmd_name}:", result.stderr)
    except subprocess.CalledProcessError as e:
        print(f"Error while executing {cmd}:")
        print(f"Return Code: {e.returncode}")
        print(f"STDOUT: {e.stdout}")
        print(f"STDERR: {e.stderr}")
        raise


def analysis(tardir: str, casacmd='casa', skip=True):
    """
    Run the analysis of the QSO data.

    Args:
        tardir (str): Directory containing the `*.asdm.sdm.tar` files.
        casacmd (str): CASA command. Default is 'casa'.
        skip (bool): Skip the analysis if the output directory exists. Default is True.
    """
    asdm_files = [file for file in os.listdir(f'{tardir}') if file.endswith('.asdm.sdm.tar')]

    for asdm_file in asdm_files:
        cmd = "sys.path.append('.');" + \
            "from almaqso._analysis import _analysis;" + \
            f"_analysis('{asdm_file}', skip={skip}," + \
            f"casacmd='{casacmd}')"
        _run_casa_script(casacmd, cmd, asdm_file)



import almaqa2csg as csg
import analysisUtils as aU
import os
import sys
import numpy as np
import glob
import subprocess


def _run_casa_cmd(casa, cmd):
    try:
        result = subprocess.run(
            [casa, '--nologger', '--nogui', '-c', f'"{cmd}"'],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True
        )
        print(f"STDOUT for {cmd}:", result.stdout)
        print(f"STDERR for {cmd}:", result.stderr)
    except subprocess.CalledProcessError as e:
        print(f"Error while executing {cmd}:")
        print(f"Return Code: {e.returncode}")
        print(f"STDOUT: {e.stdout}")
        print(f"STDERR: {e.stderr}")
        raise


class QSOanalysis():
    def __init__(self, tarfilename, casacmd='casa', casacmdforuvfit='casa',
                 spacesave=False, workingDir=None):
        self.tarfilename = tarfilename
        self.workingDir = workingDir
        self.spacesave = spacesave
        self.casacmd = casacmd
        self.casacmdforuvfit = casacmdforuvfit

        self.projID = tarfilename.split('_uid___')[0]
        self.asdmname = 'uid___' + \
            (tarfilename.split('_uid___')[1]).replace('.asdm.sdm.tar', '')

        try:
            print(f'analysisUtils of {aU.version()} will be used.')
        except Exception:
            raise Exception('analysisUtils is not found')

    def writelog(self, content=''):
        os.system('mkdir -p log')
        os.system('touch ./log/'+self.asdmname+'.analysis.v2.log')
        os.system('echo "'+content+'" >> '+'./log/' +
                  self.asdmname+'.analysis.v2.log')

    # step0: untar & make working dir
    def intial_proc(self, forcerun=False):
        if self.workingDir is not None:
            os.chdir(self.workingDir)

        if os.path.exists(self.tarfilename):
            os.makedirs(self.asdmname)
            os.system('cp '+self.tarfilename+' '+self.asdmname+'/')
            os.chdir(self.asdmname)
            os.system('tar -xf '+self.tarfilename)

        elif os.path.exists(self.tarfilename+'.gz'):
            os.makedirs(self.asdmname)
            os.system('cp '+self.tarfilename+'.gz '+self.asdmname+'/')
            os.chdir(self.asdmname)
            os.system('gzip -d '+self.tarfilename+'.gz')
            os.system('tar -xvf '+self.tarfilename)

        elif os.path.exists(self.asdmname):
            os.chdir(self.asdmname)

            if os.path.exists(self.tarfilename):
                os.system('tar -xvf '+self.tarfilename)

            elif os.path.exists(self.tarfilename+'.gz'):
                os.system('gzip -d '+self.tarfilename+'.gz')
                os.system('tar -xvf '+self.tarfilename)

        else:
            raise Exception('You may need to download data.')

        self.writelog('step0:OK')

    # step1: importasdm
    def importasdm(self):
        asdmfile = glob.glob(
            f'{os.getcwd()}/{os.path.basename(self.projID)}/*/*/*/raw/*')[0]
        visname = (os.path.basename(asdmfile)).replace('.asdm.sdm', '.ms')

        kw_importasdm = {
            'asdm': asdmfile,
            'vis': visname,
            'asis': 'Antenna Station Receiver Source CalAtmosphere CalWVR CorrelatorMode SBSummary',
            'bdfflags': True,
            'lazy': True,
            'flagbackup': False,
        }

        from casatasks import importasdm
        os.system('rm -rf '+kw_importasdm['vis'])
        importasdm(**kw_importasdm)

        try:
            self.spws = aU.getScienceSpws(vis=visname).split(",")
            os.system('mkdir -p tempfiles')
            np.save('tempfiles/spws.npy', np.array(self.spws))

        except Exception:
            self.spws = np.load('tempfiles/spws.npy')

        self.asdmfile = asdmfile
        self.visname = visname

        self.writelog('step1:OK')

    # step2: generate calib script
    def gen_calib_script(self):
        try:
            refant = aU.commonAntennas(self.visname)
            vis = self.visname
        except Exception:
            refant = aU.commonAntennas(self.visname + '.split')
            vis = self.visname + '.split'

        if os.path.exists(
                f'./log/{self.visname}.scriptForCalibration.py'
            ):
            print('Calibration script already exists.')
            os.system(f'cp ./log/{self.visname}.scriptForCalibration.py ./')
        else:
            print('Calibration script will be generated.')
            csg.generateReducScript(
                    msNames = vis,
                    refant = refant[0],
                    corrAntPos = False,
                    useCalibratorService = False,
                    useLocalAlmaHelper = False
            )
            os.makedirs('./log', exist_ok=True)
            os.system(f'cp {self.visname}.scriptForCalibration.py ./log/')

        self.refant = refant
        self.dish_diameter = aU.almaAntennaDiameter(refant[0])

        self.writelog('step2:OK')

    # step3: remove TARGET observations
    def remove_target(self):
        IntentListASDM = aU.getIntentsFromASDM(self.asdmfile)

        IntentList = []
        for intfield in list(IntentListASDM):
            IntentList = IntentList + IntentListASDM[intfield]

        listOfIntents_init = (np.unique(IntentList)[
                              np.unique(IntentList) != 'OBSERVE_TARGET'])

        os.system('rm -rf '+self.visname+'.org')
        os.system('mv '+self.visname+' '+self.visname+'.org')
        kw_mstransform = {
            'vis': self.visname+'.org',
            'outputvis': self.visname,
            'datacolumn': 'all',
            'intent': '*,'.join(listOfIntents_init)+'*',
            'keepflags': True
        }

        from casatasks import mstransform
        os.system('rm -rf '+kw_mstransform['outputvis'])
        os.system('rm -rf '+kw_mstransform['outputvis']+'.flagversions')
        mstransform(**kw_mstransform)

        self.writelog('step3:OK')

    # step4: do Calibration
    def doCalib(self):
        cmdfile = self.visname + '.scriptForCalibration.py'

        checksteps = open(cmdfile, 'r')
        syscalcheck = checksteps.readlines().copy()[21]
        checksteps.close()

        f = open(cmdfile.replace('.py', '.part.py'), 'w')
        if syscalcheck.split(':')[1].split("'")[1] == \
                'Application of the bandpass and gain cal tables':
            f.write(
                'mysteps = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16]'+'\n')
        else:
            f.write(
                'mysteps = [0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17]'+'\n')
        f.write('applyonly = True'+'\n')
        f.write('execfile('+'"'+cmdfile+'"'+',globals())'+'\n')
        f.close()

        cmd = '"' + \
            'execfile('+"'"+cmdfile.replace('.py',
                                            '.part.py')+"'"+')' + '"'
        os.system(self.casacmd+' --nologger --nogui -c '+cmd)

        vis = self.visname + '.split'

        fields = np.unique(aU.getCalibrators(vis=vis))
        self.fields = []

        for field in fields:
            if field[0] == 'J':
                self.fields.append(field)

        self.beamsize = aU.estimateSynthesizedBeam(vis)
        from casatools import synthesisutils
        su = synthesisutils()

        if self.dish_diameter > 10.:
            # 12m
            self.imsize = su.getOptimumSize(int(120./self.beamsize*5))
        else:
            # 7m
            self.imsize = su.getOptimumSize(int(180./self.beamsize*5))

        self.cell = '{:.3f}'.format(self.beamsize/5) + 'arcsec'

        self.writelog('step4:OK')



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
sys.path.append(os.path.abspath('..'))

project = 'ALMAQSO'
copyright = '2024, Yuki Yoshimura, skrbcr'
author = 'Yuki Yoshimura, skrbcr'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.autosummary',
]

napoleon_google_docstring = True
autosummary_generate = True

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

add_module_names = False

autodoc_member_order = 'bysource'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['../_static']



import sys
import numpy as np
import json
from almaqso.QSOquery import QSOquery


def download_archive(band: int, jfilename: str):
    """
    Download the archive data from ALMA Science Archive.

    Args:
        band (int): Band number to download.
        jfilename (str): JSON file name.
    """
    try:
        with open(jfilename, 'r') as f:
            jdict = json.load(f)
    except FileNotFoundError:
        print(f'ERROR: File "{jfilename}" not found')
        sys.exit(1)
    except json.JSONDecodeError as e:
        print(f'Error: Failed to parse JSON file "{jfilename}" (reason: {e}).')
        sys.exit(1)

    cals = []
    for i in range(len(jdict)):
        cals.append(jdict[i]['names'][0]['name'])

    cals = np.unique(cals)
    print(f'Shape of cals: {cals.shape}')

    for i, sname in enumerate(cals):
        try:
            print(f'[{i+1}/{cals.shape[0]}] {sname} -> start')
            obj = QSOquery(
                sname,
                band=band,
                replaceNAOJ=True,
                only12m=True,
                onlyFDM=True
            )
            obj.get_data_urls(almaquery=False)
            # np.save(f'./urls/{sname}.B{band}.npy', obj.url_list)
            obj.download()
            print(f'Totla size({sname} B{band}): {obj.total_size} GB')
        except Exception as e:
            print(f'ERROR: {sname} -> failed (reason: {e}) and skipped')



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='almaqso',
    use_scm_version=True,
    description='ALMA QSO analysis',
    long_description=long_description,
    long_description_content_type='text/markdown',  # 使用するフォーマットを指定
    packages=find_packages(),
    install_requires=[
        'astropy',
        'numpy',
        'pandas',
        'pyvo',
        'scipy',
        'matplotlib',
        'astroquery'
    ],
)



import numpy as np
import os
from concurrent.futures import ThreadPoolExecutor
from logging import StreamHandler, Formatter, INFO, getLogger
import logging


class QSOquery:
    """
    Query the ALMA Science Archive for the QSO data.

    Attributes:
        sname (str): Source name.
        band (str): Band number.
        almaurl (str): ALMA Science Archive URL.
        download_d (str): Download directory.
        replaceNAOJ (bool): Replace NAOJ URL.
        only12m (bool): Only 12m data.
        onlyFDM (bool): Only FDM data.
    """

    def __init__(self, sname, band='4',
                 almaurl='https://almascience.nao.ac.jp',
                 download_d='./',
                 replaceNAOJ=False, only12m=False, onlyFDM=False):
        """
        Initialize the class.

        Args:
            sname (str): Source name.
            band (str): Band number.
            almaurl (str): ALMA Science Archive URL. Default is 'https://almascience.nao.ac.jp'.
            download_d (str): Download directory. Default is './'.
            replaceNAOJ (bool): Replace NAOJ URL. Default is False.
            only12m (bool): Only 12m data. Default is False.
            onlyFDM (bool): Only FDM data. Default is False.
        """
        from astroquery.alma import Alma
        self.sname = sname
        self.band = band
        self.almaurl = almaurl
        self.myAlma = Alma()
        self.myAlma.archive_url = almaurl
        self.download_d = download_d
        self.replaceNAOJ = replaceNAOJ
        self.only12m = only12m
        self.onlyFDM = onlyFDM

    def queryALMA(self, almaquery=True):
        """
        Query ALMA data using TAP service or myAlma interface.

        Args:
            almaquery (bool): If True, use self.myAlma for querying. Otherwise, use pyvo TAP service.

        Returns:
            pandas.DataFrame: Filtered query results.
        """
        # SQL query construction
        query = f"""
            SELECT *
            FROM ivoa.obscore
            WHERE target_name = '{self.sname}'
              AND band_list = '{self.band}'
              AND data_rights = 'Public'
        """

        # Perform query using myAlma or pyvo TAP service
        if almaquery:
            ret = self.myAlma.query_tap(query).to_table().to_pandas()
        else:
            from pyvo.dal import TAPService
            # ALMA TAP service initialization
            service = TAPService(self.almaurl + "/tap")
            ret = service.search(query).to_table().to_pandas()

        # Apply filters based on conditions
        if self.only12m:
            ret = ret[ret['antenna_arrays'].str.contains('DV|DA')]

        if self.onlyFDM:
            ret = ret[ret['velocity_resolution'] < 50000]

        return ret

    def get_data_urls(self, almaquery=True):
        """
        Retrieve data URLs and total size for ALMA observations.

        Args:
            almaquery (bool): If True, query data using myAlma. Otherwise, use alternative methods.
        """
        logging.info("Starting ALMA data query...")
        rlist = self.queryALMA(almaquery=almaquery)
        mous_list = np.unique(rlist['member_ous_uid'])

        total_size = 0.0
        url_list = []

        for id, mous in enumerate(mous_list):
            uid_url_table = self.myAlma.get_data_info(mous)

            # Filter ASDM data
            url_size = [
                [url, size] for url, size in zip(
                    uid_url_table['access_url'],
                    uid_url_table['content_length']
                )
                if '.asdm.sdm.tar' in url
            ]

            if url_size:
                asdm_size = np.sum(
                    [float(size) for _, size in url_size]
                ) / 1024 / 1024 / 1024
                url_list.extend(url_size)
                print(f"[{id + 1}/{len(mous_list)}] {asdm_size:.2f} GB")
                total_size += asdm_size
            else:
                print(f"[{id + 1}/{len(mous_list)}] -> skipped (may be SV)")

        # Convert URL list to numpy array
        url_list = np.array(url_list)

        # Store results in class attributes
        self.rlist = rlist
        self.total_size = total_size
        self.url_list = url_list

        logging.info(f"Total data size: {total_size:.2f} GB")

    def wget_f(self, num):
        getLogger().info("%s start", num)
        if self.replaceNAOJ:
            download_url = (self.url_list[num][0]).replace(
                self.almaurl, "https://almascience.nao.ac.jp")
        else:
            download_url = self.url_list[num][0]
        print('wget -q --no-check-certificate -P ' +
              self.download_d+' '+download_url)
        os.system('wget -q --no-check-certificate -P ' +
                  self.download_d+' '+download_url)
        getLogger().info("%s end", num)

    def init_logger(self):
        handler = StreamHandler()
        handler.setLevel(INFO)
        handler.setFormatter(
            Formatter("[%(asctime)s] [%(threadName)s] %(message)s"))
        logger = getLogger()
        logger.addHandler(handler)
        logger.setLevel(INFO)

    def download(self):
        nFiles = self.url_list.shape[0]
        self.init_logger()
        getLogger().info("main start")
        with ThreadPoolExecutor(
                max_workers=min(nFiles, 5),
                thread_name_prefix="thread"
        ) as executor:
            for i in range(nFiles):
                executor.submit(self.wget_f, i)
            getLogger().info("submit end")
        getLogger().info("main end")
